/* *****************************************************************************
Copyright (c) 2016-2017, The Regents of the University of California (Regents).
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the FreeBSD Project.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS 
PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, 
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

*************************************************************************** */

// Written: mhgardner

#include <QComboBox>
#include <QDebug>
#include <QDoubleSpinBox>
#include <QHBoxLayout>
#include <QJsonValue>
#include <QJsonArray>
#include <QJsonObject>
#include <QLabel>
#include <QSpinBox>
#include <QString>
#include <QVBoxLayout>
#include <RandomVariablesContainer.h>
#include <StochasticWindModel/include/StochasticWindInput.h>
#include <StochasticWindModel/include/WittigSinha.h>


StochasticWindInput::StochasticWindInput(
    RandomVariablesContainer* random_variables, QWidget* parent)
    : SimCenterAppWidget(parent),
      rvInputWidget(random_variables) {
  // Construct required layouts
  QVBoxLayout* layout = new QVBoxLayout();
  QHBoxLayout* model_layout = new QHBoxLayout();
  parametersLayout = new QHBoxLayout();
  
  // Create label and add items to combo box for model selection 
  QLabel* selection_label = new QLabel(tr("Stochastic Loading Model"));
  //selection_label->setStyleSheet("font-weight: bold");
  modelSelection = new QComboBox();
  modelSelection->setObjectName("StochasticLoadingModel");
  modelSelection->addItem(tr("Wittig & Sinha (1975)"));
  stochasticModel = new WittigSinha(rvInputWidget, this);
  
  // Add widgets to layouts and layouts to this

  model_layout->addWidget(modelSelection);
  model_layout->addStretch();
  parametersLayout->addWidget(stochasticModel);
  parametersLayout->addStretch();
  layout->addWidget(selection_label);
  layout->addLayout(model_layout);
  layout->addLayout(parametersLayout);
  layout->addStretch();
  this->setLayout(layout);

  // Connect model selection slot
  connect(modelSelection,
          QOverload<const QString&>::of(&QComboBox::currentIndexChanged), this,
          &StochasticWindInput::modelSelectionChanged);
}


StochasticWindInput::~StochasticWindInput() {

}

bool StochasticWindInput::outputToJSON(QJsonObject& jsonObject) {
  bool result = false;
  if (stochasticModel != NULL) 
    return stochasticModel->outputToJSON(jsonObject);
  else {
    qDebug() << "StocashicWindInput::outputToJSON - NULL model";
  }

  return result;
}

bool StochasticWindInput::inputFromJSON(QJsonObject& jsonObject) {

  bool result = false;
  if (stochasticModel != NULL) 
    return stochasticModel->inputFromJSON(jsonObject);
  else {
    qDebug() << "StocashicWindInput::outputToJSON - NULL model";
  }

  return result;
}

bool StochasticWindInput::outputAppDataToJSON(QJsonObject& jsonObject) {

  bool result = false;
  if (stochasticModel != NULL) 
    return stochasticModel->outputAppDataToJSON(jsonObject);
  else {
    qDebug() << "StocashicWindInput::outputToJSON - NULL model";
  }
  return result;
}


bool StochasticWindInput::inputAppDataFromJSON(QJsonObject& jsonObject) {

  QString appName;
  appName = jsonObject.value("Application").toString();
  if (appName == "StochasticWindInput-WittigSinha1975") {  

    this->modelSelectionChanged(QString("Wittig & Sinha (1975)"));
    stochasticModel->inputAppDataFromJSON(jsonObject); // no check for NULL as cannot be if i can write code!
  } 
  
  else {
    QString message = QString("StocashicWindInput::inputAppDataFromJSON - unknown application string: ") + appName;
    //    qDebug() << message;
    emit errorMessage(message);
  }

  return true;
}

void StochasticWindInput::modelSelectionChanged(const QString& model) {

  // Switch the model description and form layout based on model selection
  SimCenterAppWidget *nextModel = NULL;
  if (model == "Wittig & Sinha (1975)") {
    nextModel = new WittigSinha(rvInputWidget, this);
  } else {
    qDebug() << "ERROR: In StochasticWindInput::modelSelectionChanged: "
                "Unknown selection: "
             << model << "\n";
  }
  
  if (nextModel != NULL) {
    if (stochasticModel != NULL) {
      parametersLayout->replaceWidget(stochasticModel, nextModel);
      delete stochasticModel;
      stochasticModel = nextModel;
    }
  }
}

void StochasticWindInput::errorMessage(QString message) {
  emit sendErrorMessage(message);
}
